package Gscan2pdf::Helpers;

use strict;
use warnings;
use IPC::Cmd qw(can_run);

my $magick_command;

sub get_imagemagick_command {
    if ( not defined $magick_command ) {
        if ( can_run('magick') ) {
            $magick_command = 'magick';
        }
        else {
            $magick_command = 'convert';
        }
    }
    return $magick_command;
}

1;
